@@ -5,6 +5,7 @@
 #include <time.h>
 #include <stddef.h>
 #include <stdint.h>
+#include <limits.h>
 
 #include "outputs.h"
 
@@ -30,6 +31,7 @@ enum sort_type {
   S_VPATH,
   S_POS,
   S_SHUFFLE_POS,
+  S_DATE_RELEASED,
 };
 
 #define Q_F_BROWSE (1 << 15)
@@ -60,17 +62,13 @@ enum query_type {
 #define ARTWORK_UNKNOWN   0
 #define ARTWORK_NONE      1
 #define ARTWORK_EMBEDDED  2
-#define ARTWORK_OWN       3
-#define ARTWORK_DIR       4
-#define ARTWORK_PARENTDIR 5
-#define ARTWORK_SPOTIFY   6
-#define ARTWORK_HTTP      7
 
 #define DB_ADMIN_SCHEMA_VERSION_MAJOR "schema_version_major"
 #define DB_ADMIN_SCHEMA_VERSION_MINOR "schema_version_minor"
 #define DB_ADMIN_SCHEMA_VERSION "schema_version"
 #define DB_ADMIN_QUEUE_VERSION "queue_version"
 #define DB_ADMIN_DB_UPDATE "db_update"
+#define DB_ADMIN_DB_MODIFIED "db_modified"
 #define DB_ADMIN_START_TIME "start_time"
 #define DB_ADMIN_LASTFM_SESSION_KEY "lastfm_sk"
 #define DB_ADMIN_SPOTIFY_REFRESH_TOKEN "spotify_refresh_token"
@@ -93,9 +91,12 @@ struct query_params {
 
   char *having;
   char *order;
+  char *group;
 
   char *filter;
 
+  int with_disabled;
+
   /* Query results, filled in by query_start */
   int results;
 
@@ -121,6 +122,8 @@ enum media_kind {
   MEDIA_KIND_TVSHOW = 64,
 };
 
+#define MEDIA_KIND_ALL USHRT_MAX
+
 const char *
 db_media_kind_label(enum media_kind media_kind);
 
@@ -140,12 +143,16 @@ db_data_kind_label(enum data_kind data_kind);
 
 /* Note that fields marked as integers in the metadata map in filescanner_ffmpeg must be uint32_t here */
 struct media_file_info {
+  uint32_t id;
+
   char *path;
-  uint32_t index;
+  char *virtual_path;
   char *fname;
+  uint32_t directory_id; /* Id of directory */
   char *title;
   char *artist;
   char *album;
+  char *album_artist;
   char *genre;
   char *comment;
   char *type;            /* daap.songformat */
@@ -157,9 +164,11 @@ struct media_file_info {
 
   uint32_t bitrate;
   uint32_t samplerate;
+  uint32_t channels;
   uint32_t song_length;
   int64_t file_size;
   uint32_t year;         /* TDRC */
+  uint32_t date_released;
 
   uint32_t track;        /* TRCK */
   uint32_t total_tracks;
@@ -167,44 +176,44 @@ struct media_file_info {
   uint32_t disc;         /* TPOS */
   uint32_t total_discs;
 
+  uint32_t bpm;          /* TBPM */
+  uint32_t compilation;
+  uint32_t artwork;
+  uint32_t rating;
+
+  uint32_t play_count;
+  uint32_t skip_count;
+  uint32_t seek;
+
+  uint32_t data_kind;    /* dmap.datakind (asdk) */
+  uint32_t media_kind;
+  uint32_t item_kind;    /* song or movie */
+
+  char *description;     /* daap.songdescription */
+
+  uint32_t db_timestamp;
   uint32_t time_added;   /* FIXME: time_t */
   uint32_t time_modified;
   uint32_t time_played;
-
-  uint32_t play_count;
-  uint32_t seek;
-  uint32_t rating;
-  uint32_t db_timestamp;
+  uint32_t time_skipped;
 
   uint32_t disabled;
-  uint32_t bpm;          /* TBPM */
 
-  uint32_t id;
-
-  char *description;     /* daap.songdescription */
+  uint64_t sample_count; //TODO [unused] sample count is never set and therefor always 0
   char *codectype;       /* song.codectype, 4 chars max (32 bits) */
 
-  uint32_t item_kind;    /* song or movie */
-  uint32_t data_kind;    /* dmap.datakind (asdk) */
-  uint64_t sample_count; //TODO [unused] sample count is never set and therefor always 0
-  uint32_t compilation;
-  char artwork;
+  uint32_t idx;
 
-  /* iTunes 5+ */
-  uint32_t contentrating;
+  uint32_t has_video;    /* iTunes 6.0.2 */
+  uint32_t contentrating;/* iTunes 5+ */
 
-  /* iTunes 6.0.2 */
-  uint32_t has_video;
   uint32_t bits_per_sample;
 
-  uint32_t media_kind;
-  uint32_t tv_episode_sort;
-  uint32_t tv_season_num;
   char *tv_series_name;
   char *tv_episode_num_str; /* com.apple.itunes.episode-num-str, used as a unique episode identifier */
   char *tv_network_name;
-
-  char *album_artist;
+  uint32_t tv_episode_sort;
+  uint32_t tv_season_num;
 
   int64_t songartistid;
   int64_t songalbumid;
@@ -212,35 +221,30 @@ struct media_file_info {
   char *title_sort;
   char *artist_sort;
   char *album_sort;
-  char *composer_sort;
   char *album_artist_sort;
-
-  char *virtual_path;
-
-  uint32_t directory_id; /* Id of directory */
-  uint32_t date_released;
-
-  uint32_t skip_count;
-  uint32_t time_skipped;
+  char *composer_sort;
 };
 
 #define mfi_offsetof(field) offsetof(struct media_file_info, field)
 
+/* Keep in sync with pl_type_label[] */
 /* PL_SPECIAL value must be in sync with type value in Q_PL* in db_init.c */
 enum pl_type {
   PL_SPECIAL = 0,
   PL_FOLDER = 1,
   PL_SMART = 2,
   PL_PLAIN = 3,
+  PL_RSS  = 4,
   PL_MAX,
 };
 
+const char *
+db_pl_type_label(enum pl_type pl_type);
+
 struct playlist_info {
   uint32_t id;           /* integer id (miid) */
   char *title;           /* playlist name as displayed in iTunes (minm) */
   enum pl_type type;     /* see PL_ types */
-  uint32_t items;        /* number of items (mimc) */
-  uint32_t streams;      /* number of internet streams */
   char *query;           /* where clause if type 1 (MSPS) */
   uint32_t db_timestamp; /* time last updated */
   uint32_t disabled;
@@ -250,8 +254,12 @@ struct playlist_info {
   char *virtual_path;    /* virtual path of underlying playlist */
   uint32_t parent_id;    /* Id of parent playlist if the playlist is nested */
   uint32_t directory_id; /* Id of directory */
-  char *query_order;     /* order by clause if it is a smart playlist */
-  int32_t query_limit;   /* limit if it is a smart playlist */
+  char *query_order;     /* order by clause, used by e.g. a smart playlists */
+  int32_t query_limit;   /* limit, used by e.g. smart playlists */
+  uint32_t media_kind;
+  char *artwork_url;     /* optional artwork */
+  uint32_t items;        /* number of items (mimc) */
+  uint32_t streams;      /* number of internet streams */
 };
 
 #define pli_offsetof(field) offsetof(struct playlist_info, field)
@@ -260,8 +268,6 @@ struct db_playlist_info {
   char *id;
   char *title;
   char *type;
-  char *items;
-  char *streams;
   char *query;
   char *db_timestamp;
   char *disabled;
@@ -273,6 +279,10 @@ struct db_playlist_info {
   char *directory_id;
   char *query_order;
   char *query_limit;
+  char *media_kind;
+  char *artwork_url;
+  char *items;
+  char *streams;
 };
 
 #define dbpli_offsetof(field) offsetof(struct db_playlist_info, field)
@@ -308,10 +318,13 @@ struct db_group_info {
 struct db_media_file_info {
   char *id;
   char *path;
+  char *virtual_path;
   char *fname;
+  char *directory_id;
   char *title;
   char *artist;
   char *album;
+  char *album_artist;
   char *genre;
   char *comment;
   char *type;
@@ -325,6 +338,7 @@ struct db_media_file_info {
   char *song_length;
   char *file_size;
   char *year;
+  char *date_released;
   char *track;
   char *total_tracks;
   char *disc;
@@ -334,14 +348,17 @@ struct db_media_file_info {
   char *artwork;
   char *rating;
   char *play_count;
+  char *skip_count;
   char *seek;
   char *data_kind;
+  char *media_kind;
   char *item_kind;
   char *description;
+  char *db_timestamp;
   char *time_added;
   char *time_modified;
   char *time_played;
-  char *db_timestamp;
+  char *time_skipped;
   char *disabled;
   char *sample_count;
   char *codectype;
@@ -349,8 +366,6 @@ struct db_media_file_info {
   char *has_video;
   char *contentrating;
   char *bits_per_sample;
-  char *album_artist;
-  char *media_kind;
   char *tv_episode_sort;
   char *tv_season_num;
   char *tv_series_name;
@@ -361,13 +376,9 @@ struct db_media_file_info {
   char *title_sort;
   char *artist_sort;
   char *album_sort;
-  char *composer_sort;
   char *album_artist_sort;
-  char *virtual_path;
-  char *directory_id;
-  char *date_released;
-  char *skip_count;
-  char *time_skipped;
+  char *composer_sort;
+  char *channels;
 };
 
 #define dbmfi_offsetof(field) offsetof(struct db_media_file_info, field)
@@ -412,6 +423,7 @@ enum directory_ids {
 struct directory_info {
   uint32_t id;
   char *virtual_path;
+  char *path;
   uint32_t db_timestamp;
   uint32_t disabled;
   uint32_t parent_id;
@@ -424,8 +436,7 @@ struct directory_enum {
   void *stmt;
 };
 
-struct db_queue_item
-{
+struct db_queue_item {
   /* A unique id for this queue item. If the same item appears multiple
      times in the queue each corresponding queue item has its own id. */
   uint32_t id;
@@ -433,18 +444,16 @@ struct db_queue_item
   /* Id of the file/item in the files database */
   uint32_t file_id;
 
-  /* Length of the item in ms */
-  uint32_t song_length;
+  uint32_t pos;
+  uint32_t shuffle_pos;
 
   /* Data type of the item */
   enum data_kind data_kind;
   /* Media type of the item */
   enum media_kind media_kind;
 
-  uint32_t seek;
-
-  uint32_t pos;
-  uint32_t shuffle_pos;
+  /* Length of the item in ms */
+  uint32_t song_length;
 
   char *path;
   char *virtual_path;
@@ -469,12 +478,30 @@ struct db_queue_item
   char *artwork_url;
 
   uint32_t queue_version;
+
+  char *composer;
+
+  char *type;
+  uint32_t bitrate;
+  uint32_t samplerate;
+  uint32_t channels;
+
+  int64_t songartistid;
+
+  /* Not saved in queue table */
+  uint32_t seek;
 };
 
+#define qi_offsetof(field) offsetof(struct db_queue_item, field)
+
 struct db_queue_add_info
 {
   int queue_version;
+  int start_pos;
   int pos;
+  int shuffle_pos;
+  int count;
+  int new_item_id;
 };
 
 char *
@@ -504,12 +531,6 @@ free_query_params(struct query_params *qp, int content_only);
 void
 free_queue_item(struct db_queue_item *queue_item, int content_only);
 
-void
-unicode_fixup_mfi(struct media_file_info *mfi);
-
-void
-fixup_tags_mfi(struct media_file_info *mfi);
-
 /* Maintenance and DB hygiene */
 void
 db_hook_post_scan(void);
@@ -541,7 +562,7 @@ int
 db_query_fetch_file(struct query_params *qp, struct db_media_file_info *dbmfi);
 
 int
-db_query_fetch_pl(struct query_params *qp, struct db_playlist_info *dbpli, int with_itemcount);
+db_query_fetch_pl(struct query_params *qp, struct db_playlist_info *dbpli);
 
 int
 db_query_fetch_group(struct query_params *qp, struct db_group_info *dbgri);
@@ -557,14 +578,23 @@ db_query_fetch_string_sort(struct query_params *qp, char **string, char **sortst
 
 /* Files */
 int
-db_files_get_count(void);
+db_files_get_count(uint32_t *nitems, uint32_t *nstreams, const char *filter);
 
 void
 db_file_inc_playcount(int id);
 
+void
+db_file_inc_playcount_byplid(int id, bool only_unplayed);
+
+void
+db_file_inc_playcount_bysongalbumid(int64_t id, bool only_unplayed);
+
 void
 db_file_inc_skipcount(int id);
 
+void
+db_file_reset_playskip_count(int id);
+
 void
 db_file_ping(int id);
 
@@ -630,7 +660,7 @@ db_filecount_get(struct filecount_info *fci, struct query_params *qp);
 
 /* Playlists */
 int
-db_pl_get_count(void);
+db_pl_get_count(uint32_t *nitems);
 
 void
 db_pl_ping(int id);
@@ -644,6 +674,9 @@ db_pl_ping_items_bymatch(const char *path, int id);
 int
 db_pl_id_bypath(const char *path);
 
+struct playlist_info *
+db_pl_fetch_byid(int id);
+
 struct playlist_info *
 db_pl_fetch_bypath(const char *path);
 
@@ -654,7 +687,10 @@ struct playlist_info *
 db_pl_fetch_bytitlepath(const char *title, const char *path);
 
 int
-db_pl_add(struct playlist_info *pli, int *id);
+db_pl_add(struct playlist_info *pli);
+
+int
+db_pl_update(struct playlist_info *pli);
 
 int
 db_pl_add_item_bypath(int plid, const char *path);
@@ -665,9 +701,6 @@ db_pl_add_item_byid(int plid, int fileid);
 void
 db_pl_clear_items(int id);
 
-int
-db_pl_update(struct playlist_info *pli);
-
 void
 db_pl_delete(int id);
 
@@ -693,7 +726,10 @@ db_group_persistentid_byid(int id, int64_t *persistentid);
 
 /* Directories */
 int
-db_directory_id_byvirtualpath(char *virtual_path);
+db_directory_id_byvirtualpath(const char *virtual_path);
+
+int
+db_directory_id_bypath(const char *path);
 
 int
 db_directory_enum_start(struct directory_enum *de);
@@ -705,7 +741,7 @@ void
 db_directory_enum_end(struct directory_enum *de);
 
 int
-db_directory_addorupdate(char *virtual_path, int disabled, int parent_id);
+db_directory_addorupdate(char *virtual_path, char *path, int disabled, int parent_id);
 
 void
 db_directory_ping_bymatch(char *virtual_path);
@@ -726,7 +762,6 @@ db_pairing_add(struct pairing_info *pi);
 int
 db_pairing_fetch_byguid(struct pairing_info *pi);
 
-#ifdef HAVE_SPOTIFY_H
 /* Spotify */
 void
 db_spotify_purge(void);
@@ -736,7 +771,6 @@ db_spotify_pl_delete(int id);
 
 void
 db_spotify_files_delete(void);
-#endif
 
 /* Admin */
 int
@@ -748,14 +782,14 @@ db_admin_setint(const char *key, int value);
 int
 db_admin_setint64(const char *key, int64_t value);
 
-char *
-db_admin_get(const char *key);
+int
+db_admin_get(char **value, const char *key);
 
 int
-db_admin_getint(const char *key);
+db_admin_getint(int *intval, const char *key);
 
-int64_t
-db_admin_getint64(const char *key);
+int
+db_admin_getint64(int64_t *int64val, const char *key);
 
 int
 db_admin_delete(const char *key);
@@ -787,10 +821,10 @@ int
 db_queue_add_by_fileid(int id, char reshuffle, uint32_t item_id, int position, int *count, int *new_item_id);
 
 int
-db_queue_add_start(struct db_queue_add_info *queue_add_info);
+db_queue_add_start(struct db_queue_add_info *queue_add_info, int pos);
 
-void
-db_queue_add_end(struct db_queue_add_info *queue_add_info, int ret);
+int
+db_queue_add_end(struct db_queue_add_info *queue_add_info, char reshuffle, uint32_t item_id, int ret);
 
 int
 db_queue_add_item(struct db_queue_add_info *queue_add_info, struct db_queue_item *item);
@@ -850,7 +884,10 @@ int
 db_queue_reshuffle(uint32_t item_id);
 
 int
-db_queue_get_count();
+db_queue_inc_version(void);
+
+int
+db_queue_get_count(uint32_t *nitems);
 
 int
 db_queue_get_pos(uint32_t item_id, char shuffle);
