@@ -36,10 +36,13 @@
   "CREATE TABLE IF NOT EXISTS files ("			\
   "   id                 INTEGER PRIMARY KEY NOT NULL,"	\
   "   path               VARCHAR(4096) NOT NULL,"	\
+  "   virtual_path       VARCHAR(4096) DEFAULT NULL,"	\
   "   fname              VARCHAR(255) NOT NULL,"	\
+  "   directory_id       INTEGER DEFAULT 0,"		\
   "   title              VARCHAR(1024) DEFAULT NULL COLLATE DAAP,"	\
   "   artist             VARCHAR(1024) DEFAULT NULL COLLATE DAAP,"	\
   "   album              VARCHAR(1024) NOT NULL COLLATE DAAP,"		\
+  "   album_artist       VARCHAR(1024) NOT NULL COLLATE DAAP,"		\
   "   genre              VARCHAR(255) DEFAULT NULL COLLATE DAAP,"	\
   "   comment            VARCHAR(4096) DEFAULT NULL COLLATE DAAP,"	\
   "   type               VARCHAR(255) DEFAULT NULL COLLATE DAAP,"	\
@@ -53,6 +56,7 @@
   "   song_length        INTEGER DEFAULT 0,"		\
   "   file_size          INTEGER DEFAULT 0,"		\
   "   year               INTEGER DEFAULT 0,"		\
+  "   date_released      INTEGER DEFAULT 0,"		\
   "   track              INTEGER DEFAULT 0,"		\
   "   total_tracks       INTEGER DEFAULT 0,"		\
   "   disc               INTEGER DEFAULT 0,"		\
@@ -62,14 +66,17 @@
   "   artwork            INTEGER DEFAULT 0,"		\
   "   rating             INTEGER DEFAULT 0,"		\
   "   play_count         INTEGER DEFAULT 0,"		\
+  "   skip_count         INTEGER DEFAULT 0,"		\
   "   seek               INTEGER DEFAULT 0,"		\
   "   data_kind          INTEGER DEFAULT 0,"		\
+  "   media_kind         INTEGER DEFAULT 0,"		\
   "   item_kind          INTEGER DEFAULT 0,"		\
   "   description        INTEGER DEFAULT 0,"		\
+  "   db_timestamp       INTEGER DEFAULT 0,"		\
   "   time_added         INTEGER DEFAULT 0,"		\
   "   time_modified      INTEGER DEFAULT 0,"		\
   "   time_played        INTEGER DEFAULT 0,"		\
-  "   db_timestamp       INTEGER DEFAULT 0,"		\
+  "   time_skipped       INTEGER DEFAULT 0,"		\
   "   disabled           INTEGER DEFAULT 0,"		\
   "   sample_count       INTEGER DEFAULT 0,"		\
   "   codectype          VARCHAR(5) DEFAULT NULL,"	\
@@ -77,25 +84,19 @@
   "   has_video          INTEGER DEFAULT 0,"		\
   "   contentrating      INTEGER DEFAULT 0,"		\
   "   bits_per_sample    INTEGER DEFAULT 0,"		\
-  "   album_artist       VARCHAR(1024) NOT NULL COLLATE DAAP,"		\
-  "   media_kind         INTEGER NOT NULL,"		\
   "   tv_series_name     VARCHAR(1024) DEFAULT NULL COLLATE DAAP,"	\
   "   tv_episode_num_str VARCHAR(1024) DEFAULT NULL COLLATE DAAP,"	\
   "   tv_network_name    VARCHAR(1024) DEFAULT NULL COLLATE DAAP,"	\
   "   tv_episode_sort    INTEGER NOT NULL,"		\
   "   tv_season_num      INTEGER NOT NULL,"		\
-  "   songartistid       INTEGER NOT NULL,"		\
-  "   songalbumid        INTEGER NOT NULL,"		\
+  "   songartistid       INTEGER DEFAULT 0,"		\
+  "   songalbumid        INTEGER DEFAULT 0,"		\
   "   title_sort         VARCHAR(1024) DEFAULT NULL COLLATE DAAP,"	\
   "   artist_sort        VARCHAR(1024) DEFAULT NULL COLLATE DAAP,"	\
   "   album_sort         VARCHAR(1024) DEFAULT NULL COLLATE DAAP,"	\
-  "   composer_sort      VARCHAR(1024) DEFAULT NULL COLLATE DAAP,"	\
   "   album_artist_sort  VARCHAR(1024) DEFAULT NULL COLLATE DAAP,"	\
-  "   virtual_path       VARCHAR(4096) DEFAULT NULL,"	\
-  "   directory_id       INTEGER DEFAULT 0,"		\
-  "   date_released      INTEGER DEFAULT 0,"            \
-  "   skip_count         INTEGER DEFAULT 0,"            \
-  "   time_skipped       INTEGER DEFAULT 0"             \
+  "   composer_sort      VARCHAR(1024) DEFAULT NULL COLLATE DAAP,"	\
+  "   channels           INTEGER DEFAULT 0"		\
   ");"
 
 #define T_PL					\
@@ -113,7 +114,9 @@
   "   parent_id      INTEGER DEFAULT 0,"		\
   "   directory_id   INTEGER DEFAULT 0,"		\
   "   query_order    VARCHAR(1024),"			\
-  "   query_limit    INTEGER DEFAULT -1"		\
+  "   query_limit    INTEGER DEFAULT -1,"		\
+  "   media_kind     INTEGER DEFAULT 1,"		\
+  "   artwork_url    VARCHAR(4096) DEFAULT NULL"	\
   ");"
 
 #define T_PLITEMS				\
@@ -161,7 +164,8 @@
   "   virtual_path        VARCHAR(4096) NOT NULL,"		\
   "   db_timestamp        INTEGER DEFAULT 0,"			\
   "   disabled            INTEGER DEFAULT 0,"			\
-  "   parent_id           INTEGER DEFAULT 0"			\
+  "   parent_id           INTEGER DEFAULT 0,"			\
+  "   path                VARCHAR(4096) DEFAULT NULL"		\
   ");"
 
 #define T_QUEUE								\
@@ -189,23 +193,15 @@
   "   track               INTEGER DEFAULT 0,"				\
   "   disc                INTEGER DEFAULT 0,"				\
   "   artwork_url         VARCHAR(4096) DEFAULT NULL,"			\
-  "   queue_version       INTEGER DEFAULT 0"				\
+  "   queue_version       INTEGER DEFAULT 0,"				\
+  "   composer            VARCHAR(1024) DEFAULT NULL,"			\
+  "   songartistid        INTEGER NOT NULL,"				\
+  "   type                VARCHAR(8) DEFAULT NULL,"			\
+  "   bitrate             INTEGER DEFAULT 0,"				\
+  "   samplerate          INTEGER DEFAULT 0,"				\
+  "   channels            INTEGER DEFAULT 0"				\
   ");"
 
-#define TRG_GROUPS_INSERT_FILES						\
-  "CREATE TRIGGER update_groups_new_file AFTER INSERT ON files FOR EACH ROW" \
-  " BEGIN"								\
-  "   INSERT OR IGNORE INTO groups (type, name, persistentid) VALUES (1, NEW.album, NEW.songalbumid);" \
-  "   INSERT OR IGNORE INTO groups (type, name, persistentid) VALUES (2, NEW.album_artist, NEW.songartistid);" \
-  " END;"
-
-#define TRG_GROUPS_UPDATE_FILES						\
-  "CREATE TRIGGER update_groups_update_file AFTER UPDATE OF songalbumid ON files FOR EACH ROW" \
-  " BEGIN"								\
-  "   INSERT OR IGNORE INTO groups (type, name, persistentid) VALUES (1, NEW.album, NEW.songalbumid);" \
-  "   INSERT OR IGNORE INTO groups (type, name, persistentid) VALUES (2, NEW.album_artist, NEW.songartistid);" \
-  " END;"
-
 #define Q_PL1								\
   "INSERT INTO playlists (id, title, type, query, db_timestamp, path, idx, special_id)" \
   " VALUES(1, 'Library', 0, '1 = 1', 0, '', 0, 0);"
@@ -238,17 +234,17 @@
 
 
 #define Q_DIR1 \
-  "INSERT INTO directories (id, virtual_path, db_timestamp, disabled, parent_id)" \
-  " VALUES (1, '/', 0, 0, 0);"
+  "INSERT INTO directories (id, virtual_path, db_timestamp, disabled, parent_id, path)" \
+  " VALUES (1, '/', 0, 0, 0, NULL);"
 #define Q_DIR2 \
-  "INSERT INTO directories (id, virtual_path, db_timestamp, disabled, parent_id)" \
-  " VALUES (2, '/file:', 0, 0, 1);"
+  "INSERT INTO directories (id, virtual_path, db_timestamp, disabled, parent_id, path)" \
+  " VALUES (2, '/file:', 0, 0, 1, '/');"
 #define Q_DIR3 \
-  "INSERT INTO directories (id, virtual_path, db_timestamp, disabled, parent_id)" \
-  " VALUES (3, '/http:', 0, 0, 1);"
+  "INSERT INTO directories (id, virtual_path, db_timestamp, disabled, parent_id, path)" \
+  " VALUES (3, '/http:', 0, 0, 1, NULL);"
 #define Q_DIR4 \
-  "INSERT INTO directories (id, virtual_path, db_timestamp, disabled, parent_id)" \
-  " VALUES (4, '/spotify:', 0, 4294967296, 1);"
+  "INSERT INTO directories (id, virtual_path, db_timestamp, disabled, parent_id, path)" \
+  " VALUES (4, '/spotify:', 0, 4294967296, 1, NULL);"
 
 #define Q_QUEUE_VERSION			\
   "INSERT INTO admin (key, value) VALUES ('queue_version', '0');"
@@ -276,19 +272,18 @@ static const struct db_init_query db_init_table_queries[] =
     { T_DIRECTORIES, "create table directories" },
     { T_QUEUE,     "create table queue" },
 
-    { TRG_GROUPS_INSERT_FILES,    "create trigger update_groups_new_file" },
-    { TRG_GROUPS_UPDATE_FILES,    "create trigger update_groups_update_file" },
-
     { Q_PL1,       "create default playlist" },
     { Q_PL2,       "create default smart playlist 'Music'" },
     { Q_PL3,       "create default smart playlist 'Movies'" },
     { Q_PL4,       "create default smart playlist 'TV Shows'" },
     { Q_PL5,       "create default smart playlist 'Podcasts'" },
     { Q_PL6,       "create default smart playlist 'Audiobooks'" },
+
     { Q_DIR1,      "create default root directory '/'" },
     { Q_DIR2,      "create default base directory '/file:'" },
     { Q_DIR3,      "create default base directory '/http:'" },
     { Q_DIR4,      "create default base directory '/spotify:'" },
+
     { Q_QUEUE_VERSION, "initialize queue version" },
   };
 
@@ -341,6 +336,9 @@ static const struct db_init_query db_init_table_queries[] =
 #define I_FILE_DIR					\
   "CREATE INDEX IF NOT EXISTS idx_file_dir ON files(disabled, directory_id);"
 
+#define I_DATE_RELEASED                    \
+  "CREATE INDEX IF NOT EXISTS idx_date_released ON files(disabled, date_released DESC, media_kind);"
+
 #define I_PL_PATH				\
   "CREATE INDEX IF NOT EXISTS idx_pl_path ON playlists(path);"
 
@@ -390,6 +388,7 @@ static const struct db_init_query db_init_index_queries[] =
     { I_ALBUM,     "create album index" },
     { I_FILELIST,  "create filelist index" },
     { I_FILE_DIR,  "create file dir index" },
+    { I_DATE_RELEASED, "create date_released index" },
 
     { I_PL_PATH,   "create playlist path index" },
     { I_PL_DISABLED, "create playlist state index" },
@@ -409,6 +408,30 @@ static const struct db_init_query db_init_index_queries[] =
     { I_QUEUE_SHUFFLEPOS,  "create queue shuffle pos index" },
   };
 
+
+/* Triggers must be prefixed with trg_ for db_drop_triggers() to id them */
+
+#define TRG_GROUPS_INSERT										\
+  "CREATE TRIGGER trg_groups_insert AFTER INSERT ON files FOR EACH ROW"					\
+  " BEGIN"												\
+  "   INSERT OR IGNORE INTO groups (type, name, persistentid) VALUES (1, NEW.album, NEW.songalbumid);"	\
+  "   INSERT OR IGNORE INTO groups (type, name, persistentid) VALUES (2, NEW.album_artist, NEW.songartistid);"	\
+  " END;"
+
+#define TRG_GROUPS_UPDATE										\
+  "CREATE TRIGGER trg_groups_update AFTER UPDATE OF songartistid, songalbumid ON files FOR EACH ROW"	\
+  " BEGIN"												\
+  "   INSERT OR IGNORE INTO groups (type, name, persistentid) VALUES (1, NEW.album, NEW.songalbumid);"	\
+  "   INSERT OR IGNORE INTO groups (type, name, persistentid) VALUES (2, NEW.album_artist, NEW.songartistid);"	\
+  " END;"
+
+static const struct db_init_query db_init_trigger_queries[] =
+  {
+    { TRG_GROUPS_INSERT,           "create trigger trg_groups_insert" },
+    { TRG_GROUPS_UPDATE,           "create trigger trg_groups_update" },
+  };
+
+
 int
 db_init_indices(sqlite3 *hdl)
 {
@@ -433,6 +456,30 @@ db_init_indices(sqlite3 *hdl)
   return 0;
 }
 
+int
+db_init_triggers(sqlite3 *hdl)
+{
+  char *errmsg;
+  int i;
+  int ret;
+
+  for (i = 0; i < (sizeof(db_init_trigger_queries) / sizeof(db_init_trigger_queries[0])); i++)
+    {
+      DPRINTF(E_DBG, L_DB, "DB init trigger query: %s\n", db_init_trigger_queries[i].desc);
+
+      ret = sqlite3_exec(hdl, db_init_trigger_queries[i].query, NULL, NULL, &errmsg);
+      if (ret != SQLITE_OK)
+	{
+	  DPRINTF(E_FATAL, L_DB, "DB init error: %s\n", errmsg);
+
+	  sqlite3_free(errmsg);
+	  return -1;
+	}
+    }
+
+  return 0;
+}
+
 int
 db_init_tables(sqlite3 *hdl)
 {
@@ -480,6 +527,18 @@ db_init_tables(sqlite3 *hdl)
     }
 
   ret = db_init_indices(hdl);
+  if (ret < 0)
+    {
+      DPRINTF(E_FATAL, L_DB, "DB init error: failed to create indices\n");
+      return -1;
+    }
+
+  ret = db_init_triggers(hdl);
+  if (ret < 0)
+    {
+      DPRINTF(E_FATAL, L_DB, "DB init error: failed to create triggers\n");
+      return -1;
+    }
 
   return ret;
 }
