@@ -63,6 +63,7 @@
 #include "httpd_jsonapi.h"
 #include "httpd_streaming.h"
 #include "httpd_oauth.h"
+#include "httpd_artworkapi.h"
 #include "transcode.h"
 #ifdef LASTFM
 # include "lastfm.h"
@@ -79,6 +80,11 @@
   "<h1>%s</h1>\n" \
   "</body>\n</html>\n"
 
+#define HTTPD_STREAM_SAMPLE_RATE 44100
+#define HTTPD_STREAM_BPS         16
+#define HTTPD_STREAM_CHANNELS    2
+
+
 struct content_type_map {
   char *ext;
   char *ctype;
@@ -295,7 +301,7 @@ httpd_request_etag_matches(struct evhttp_request *req, const char *etag)
 
   // Add cache headers to allow client side caching
   output_headers = evhttp_request_get_output_headers(req);
-  evhttp_add_header(output_headers, "Cache-Control", "private");
+  evhttp_add_header(output_headers, "Cache-Control", "private no-cache");
   evhttp_add_header(output_headers, "ETag", etag);
 
   return false;
@@ -313,7 +319,7 @@ httpd_request_etag_matches(struct evhttp_request *req, const char *etag)
  * @return True if the given timestamp matches the request-header-value "If-Modified-Since", otherwise false
  */
 bool
-httpd_request_not_modified_since(struct evhttp_request *req, const time_t *mtime)
+httpd_request_not_modified_since(struct evhttp_request *req, time_t mtime)
 {
   struct evkeyvalq *input_headers;
   struct evkeyvalq *output_headers;
@@ -323,7 +329,7 @@ httpd_request_not_modified_since(struct evhttp_request *req, const time_t *mtime
   input_headers = evhttp_request_get_input_headers(req);
   modified_since = evhttp_find_header(input_headers, "If-Modified-Since");
 
-  strftime(last_modified, sizeof(last_modified), "%a, %d %b %Y %H:%M:%S %Z", gmtime(mtime));
+  strftime(last_modified, sizeof(last_modified), "%a, %d %b %Y %H:%M:%S %Z", gmtime(&mtime));
 
   // Return not modified, if given timestamp matches "If-Modified-Since" request header
   if (modified_since && (strcasecmp(last_modified, modified_since) == 0))
@@ -331,7 +337,7 @@ httpd_request_not_modified_since(struct evhttp_request *req, const time_t *mtime
 
   // Add cache headers to allow client side caching
   output_headers = evhttp_request_get_output_headers(req);
-  evhttp_add_header(output_headers, "Cache-Control", "private");
+  evhttp_add_header(output_headers, "Cache-Control", "private no-cache");
   evhttp_add_header(output_headers, "Last-Modified", last_modified);
 
   return false;
@@ -449,7 +455,7 @@ serve_file(struct evhttp_request *req, const char *uri)
       return;
     }
 
-  if (httpd_request_not_modified_since(req, &sb.st_mtime))
+  if (httpd_request_not_modified_since(req, sb.st_mtime))
     {
       httpd_send_reply(req, HTTP_NOTMODIFIED, NULL, NULL, HTTPD_SEND_NO_GZIP);
       return;
@@ -842,6 +848,11 @@ httpd_gen_cb(struct evhttp_request *req, void *arg)
       jsonapi_request(req, parsed);
       goto out;
     }
+  else if (artworkapi_is_request(parsed->path))
+    {
+      artworkapi_request(req, parsed);
+      goto out;
+    }
   else if (streaming_is_request(parsed->path))
     {
       streaming_request(req, parsed);
@@ -1023,6 +1034,7 @@ httpd_request_parse(struct evhttp_request *req, struct httpd_uri_parsed *uri_par
 void
 httpd_stream_file(struct evhttp_request *req, int id)
 {
+  struct media_quality quality = { HTTPD_STREAM_SAMPLE_RATE, HTTPD_STREAM_BPS, HTTPD_STREAM_CHANNELS, 0 };
   struct media_file_info *mfi;
   struct stream_ctx *st;
   void (*stream_cb)(int fd, short event, void *arg);
@@ -1092,8 +1104,9 @@ httpd_stream_file(struct evhttp_request *req, int id)
 
   if (mfi->data_kind != DATA_KIND_FILE)
     {
-      evhttp_send_error(req, 500, "Cannot stream radio station");
+      DPRINTF(E_LOG, L_HTTPD, "Could not serve '%s' to client, not a file\n", mfi->path);
 
+      evhttp_send_error(req, 500, "Cannot stream non-file content");
       goto out_free_mfi;
     }
 
@@ -1103,7 +1116,6 @@ httpd_stream_file(struct evhttp_request *req, int id)
       DPRINTF(E_LOG, L_HTTPD, "Out of memory for struct stream_ctx\n");
 
       evhttp_send_error(req, HTTP_SERVUNAVAIL, "Internal Server Error");
-
       goto out_free_mfi;
     }
   memset(st, 0, sizeof(struct stream_ctx));
@@ -1122,7 +1134,7 @@ httpd_stream_file(struct evhttp_request *req, int id)
 
       stream_cb = stream_chunk_xcode_cb;
 
-      st->xcode = transcode_setup(XCODE_PCM16_HEADER, mfi->data_kind, mfi->path, mfi->song_length, &st->size);
+      st->xcode = transcode_setup(XCODE_PCM16_HEADER, &quality, mfi->data_kind, mfi->path, mfi->song_length, &st->size);
       if (!st->xcode)
 	{
 	  DPRINTF(E_WARN, L_HTTPD, "Transcoding setup failed, aborting streaming\n");
@@ -1543,7 +1555,7 @@ httpd_basic_auth(struct evhttp_request *req, const char *user, const char *passw
 
   auth += strlen("Basic ");
 
-  authuser = b64_decode(auth);
+  authuser = (char *)b64_decode(NULL, auth);
   if (!authuser)
     {
       DPRINTF(E_LOG, L_HTTPD, "Could not decode Authentication header\n");
@@ -1679,6 +1691,14 @@ httpd_init(const char *webroot)
       goto jsonapi_fail;
     }
 
+  ret = artworkapi_init();
+  if (ret < 0)
+    {
+      DPRINTF(E_FATAL, L_HTTPD, "Artwork init failed\n");
+
+      goto artworkapi_fail;
+    }
+
   ret = oauth_init();
   if (ret < 0)
     {
@@ -1728,7 +1748,7 @@ httpd_init(const char *webroot)
     {
       DPRINTF(E_FATAL, L_HTTPD, "Could not create exit event\n");
 
-      goto event_fail;
+      goto exitev_fail;
     }
   event_add(exitev, NULL);
 
@@ -1737,7 +1757,7 @@ httpd_init(const char *webroot)
     {
       DPRINTF(E_FATAL, L_HTTPD, "Could not create HTTP server\n");
 
-      goto event_fail;
+      goto evhttpd_fail;
     }
 
   v6enabled = cfg_getbool(cfg_getsec(cfg, "general"), "ipv6");
@@ -1799,7 +1819,9 @@ httpd_init(const char *webroot)
  thread_fail:
  bind_fail:
   evhttp_free(evhttpd);
- event_fail:
+ evhttpd_fail:
+  event_free(exitev);
+ exitev_fail:
 #ifdef HAVE_EVENTFD
   close(exit_efd);
 #else
@@ -1814,6 +1836,8 @@ httpd_init(const char *webroot)
 #endif
   oauth_deinit();
  oauth_fail:
+  artworkapi_deinit();
+ artworkapi_fail:
   jsonapi_deinit();
  jsonapi_fail:
   dacp_deinit();
@@ -1877,6 +1901,7 @@ httpd_deinit(void)
   close(exit_pipe[0]);
   close(exit_pipe[1]);
 #endif
+  event_free(exitev);
   evhttp_free(evhttpd);
   event_base_free(evbase_httpd);
 }
