@@ -48,12 +48,31 @@
 #include "listener.h"
 #include "player.h"
 
-struct playlist_add_param
+#define LIBRARY_MAX_CALLBACKS 16
+
+struct library_callback_register
+{
+  library_cb cb;
+  void *arg;
+  struct event *ev;
+};
+
+struct playlist_item_add_param
 {
   const char *vp_playlist;
   const char *vp_item;
 };
 
+struct queue_item_add_param
+{
+  const char *path;
+  int position;
+  char reshuffle;
+  uint32_t item_id;
+  int *count;
+  int *new_item_id;
+};
+
 static struct commands_base *cmdbase;
 static pthread_t tid_library;
 
@@ -63,12 +82,14 @@ extern struct library_source filescanner;
 #ifdef HAVE_SPOTIFY_H
 extern struct library_source spotifyscanner;
 #endif
+extern struct library_source rssscanner;
 
 static struct library_source *sources[] = {
     &filescanner,
 #ifdef HAVE_SPOTIFY_H
     &spotifyscanner,
 #endif
+    &rssscanner,
     NULL
 };
 
@@ -95,30 +116,20 @@ static struct event *updateev;
 static unsigned int deferred_update_notifications;
 static short deferred_update_events;
 
-static bool
-handle_deferred_update_notifications(void)
-{
-  bool ret = (deferred_update_notifications > 0);
+// Stores callbacks that backends may have requested
+static struct library_callback_register library_cb_register[LIBRARY_MAX_CALLBACKS];
 
-  if (ret)
-    {
-      DPRINTF(E_DBG, L_LIB, "Database changed (%d changes)\n", deferred_update_notifications);
 
-      deferred_update_notifications = 0;
-      db_admin_setint64(DB_ADMIN_DB_UPDATE, (int64_t) time(NULL));
-    }
+/* ------------------- CALLED BY LIBRARY SOURCE MODULES -------------------- */
 
-  return ret;
-}
-
-void
-library_add_media(struct media_file_info *mfi)
+int
+library_media_save(struct media_file_info *mfi)
 {
   if (!mfi->path || !mfi->fname)
     {
       DPRINTF(E_LOG, L_LIB, "Ignoring media file with missing values (path='%s', fname='%s', data_kind='%d')\n",
 	      mfi->path, mfi->fname, mfi->data_kind);
-      return;
+      return -1;
     }
 
   if (!mfi->directory_id || !mfi->virtual_path)
@@ -128,120 +139,128 @@ library_add_media(struct media_file_info *mfi)
 	      mfi->path, mfi->directory_id, mfi->virtual_path);
     }
 
-  if (!mfi->item_kind)
-    mfi->item_kind = 2; /* music */
-  if (!mfi->media_kind)
-    mfi->media_kind = MEDIA_KIND_MUSIC; /* music */
-
-  unicode_fixup_mfi(mfi);
-  fixup_tags_mfi(mfi);
-
   if (mfi->id == 0)
-    db_file_add(mfi);
+    return db_file_add(mfi);
   else
-    db_file_update(mfi);
+    return db_file_update(mfi);
 }
 
 int
-library_queue_add(const char *path)
+library_playlist_save(struct playlist_info *pli)
 {
-  int i;
-  int ret;
-
-  DPRINTF(E_DBG, L_LIB, "Add items for path '%s' to the queue\n", path);
-
-  ret = LIBRARY_PATH_INVALID;
-  for (i = 0; sources[i] && ret == LIBRARY_PATH_INVALID; i++)
+  if (!pli->path)
     {
-      if (sources[i]->disabled || !sources[i]->queue_add)
-        {
-	  DPRINTF(E_DBG, L_LIB, "Library source '%s' is disabled or does not support queue_add\n", sources[i]->name);
-	  continue;
-	}
+      DPRINTF(E_LOG, L_LIB, "Ignoring playlist file with missing path\n");
+      return -1;
+    }
 
-      ret = sources[i]->queue_add(path);
+  if (!pli->directory_id || !pli->virtual_path)
+    {
+      // Missing informations for virtual_path and directory_id (may) lead to misplaced appearance in mpd clients
+      DPRINTF(E_WARN, L_LIB, "Playlist with missing values (path='%s', directory='%d', virtual_path='%s')\n",
+	      pli->path, pli->directory_id, pli->virtual_path);
+    }
 
-      if (ret == LIBRARY_OK)
+  if (pli->id == 0)
+    return db_pl_add(pli);
+  else
+    return db_pl_update(pli);
+}
+
+static void
+scheduled_cb(int fd, short what, void *arg)
+{
+  struct library_callback_register *cbreg = arg;
+  library_cb cb = cbreg->cb;
+  void *cb_arg = cbreg->arg;
+
+  // Must reset the register before calling back, otherwise it won't work if the
+  // callback reschedules by calling library_callback_schedule()
+  event_free(cbreg->ev);
+  memset(cbreg, 0, sizeof(struct library_callback_register));
+
+  DPRINTF(E_DBG, L_LIB, "Executing library callback to %p\n", cb);
+  cb(cb_arg);
+}
+
+int
+library_callback_schedule(library_cb cb, void *arg, struct timeval *wait, enum library_cb_action action)
+{
+  struct library_callback_register *cbreg;
+  bool replace_done;
+  int idx_available;
+  int i;
+
+  for (i = 0, idx_available = -1, replace_done = false; i < ARRAY_SIZE(library_cb_register); i++)
+    {
+      if (idx_available == -1 && library_cb_register[i].cb == NULL)
+	idx_available = i;
+
+      if (library_cb_register[i].cb != cb)
+	continue;
+
+      if (action == LIBRARY_CB_REPLACE || action == LIBRARY_CB_ADD_OR_REPLACE)
 	{
-	  DPRINTF(E_DBG, L_LIB, "Items for path '%s' from library source '%s' added to the queue\n", path, sources[i]->name);
-	  break;
+	  event_add(library_cb_register[i].ev, wait);
+	  library_cb_register[i].arg = arg;
+	  replace_done = true;
+	}
+      else if (action == LIBRARY_CB_DELETE)
+	{
+	  event_free(library_cb_register[i].ev);
+	  memset(&library_cb_register[i], 0, sizeof(struct library_callback_register));
 	}
     }
 
-  if (ret != LIBRARY_OK)
-    DPRINTF(E_LOG, L_LIB, "Failed to add items for path '%s' to the queue (%d)\n", path, ret);
+  if (action == LIBRARY_CB_REPLACE || action == LIBRARY_CB_DELETE || (action == LIBRARY_CB_ADD_OR_REPLACE && replace_done))
+    {
+      return 0; // All done
+    }
+  else if (idx_available == -1)
+    {
+      DPRINTF(E_LOG, L_LIB, "Error scheduling callback, register full (size=%d, action=%d)\n", LIBRARY_MAX_CALLBACKS, action);
+      return -1;
+    }
+
+  cbreg = &library_cb_register[idx_available];
+  cbreg->cb = cb;
+  cbreg->arg = arg;
+
+  if (!cbreg->ev)
+    cbreg->ev = evtimer_new(evbase_lib, scheduled_cb, cbreg);
+
+  CHECK_NULL(L_LIB, cbreg->ev);
+
+  event_add(cbreg->ev, wait);
+
+  DPRINTF(E_DBG, L_LIB, "Added library callback to %p (id %d), wait %ld.%06ld\n", cbreg->cb, idx_available, wait->tv_sec, wait->tv_usec);
+
+  return idx_available;
+}
+
+
+/* ---------------------- LIBRARY ABSTRACTION --------------------- */
+/*                          thread: library                         */
+
+static bool
+handle_deferred_update_notifications(void)
+{
+  time_t update_time;
+  bool ret = (deferred_update_notifications > 0);
+
+  if (ret)
+    {
+      DPRINTF(E_DBG, L_LIB, "Database changed (%d changes)\n", deferred_update_notifications);
+
+      deferred_update_notifications = 0;
+      update_time = time(NULL);
+      db_admin_setint64(DB_ADMIN_DB_UPDATE, (int64_t) update_time);
+      db_admin_setint64(DB_ADMIN_DB_MODIFIED, (int64_t) update_time);
+    }
 
   return ret;
 }
 
-int
-library_add_playlist_info(const char *path, const char *title, const char *virtual_path, enum pl_type type, int parent_pl_id, int dir_id)
-{
-  struct playlist_info *pli;
-  int plid;
-  int ret;
-
-  pli = db_pl_fetch_bypath(path);
-  if (pli)
-    {
-      DPRINTF(E_DBG, L_LIB, "Playlist found ('%s', link %s), updating\n", title, path);
-
-      plid = pli->id;
-
-      pli->type = type;
-      free(pli->title);
-      pli->title = strdup(title);
-      if (pli->virtual_path)
-	free(pli->virtual_path);
-      pli->virtual_path = safe_strdup(virtual_path);
-      pli->directory_id = dir_id;
-
-      ret = db_pl_update(pli);
-      if (ret < 0)
-	{
-	  DPRINTF(E_LOG, L_LIB, "Error updating playlist ('%s', link %s)\n", title, path);
-
-	  free_pli(pli, 0);
-	  return -1;
-	}
-
-      db_pl_clear_items(plid);
-    }
-  else
-    {
-      DPRINTF(E_DBG, L_LIB, "Adding playlist ('%s', link %s)\n", title, path);
-
-      pli = (struct playlist_info *)malloc(sizeof(struct playlist_info));
-      if (!pli)
-	{
-	  DPRINTF(E_LOG, L_LIB, "Out of memory\n");
-
-	  return -1;
-	}
-
-      memset(pli, 0, sizeof(struct playlist_info));
-
-      pli->type = type;
-      pli->title = strdup(title);
-      pli->path = strdup(path);
-      pli->virtual_path = safe_strdup(virtual_path);
-      pli->parent_id = parent_pl_id;
-      pli->directory_id = dir_id;
-
-      ret = db_pl_add(pli, &plid);
-      if ((ret < 0) || (plid < 1))
-	{
-	  DPRINTF(E_LOG, L_LIB, "Error adding playlist ('%s', link %s, ret %d, plid %d)\n", title, path, ret, plid);
-
-	  free_pli(pli, 0);
-	  return -1;
-	}
-    }
-
-  free_pli(pli, 0);
-  return plid;
-}
-
 static void
 purge_cruft(time_t start)
 {
@@ -296,6 +315,49 @@ rescan(void *arg, int *ret)
   return COMMAND_END;
 }
 
+static enum command_state
+metarescan(void *arg, int *ret)
+{
+  time_t starttime;
+  time_t endtime;
+  int i;
+
+  DPRINTF(E_LOG, L_LIB, "Library meta rescan triggered\n");
+  listener_notify(LISTENER_UPDATE);
+  starttime = time(NULL);
+
+  for (i = 0; sources[i]; i++)
+    {
+      if (!sources[i]->disabled && sources[i]->metarescan)
+	{
+	  DPRINTF(E_INFO, L_LIB, "Meta rescan library source '%s'\n", sources[i]->name);
+	  sources[i]->metarescan();
+	}
+      else
+	{
+	  DPRINTF(E_INFO, L_LIB, "Library source '%s' is disabled\n", sources[i]->name);
+	}
+    }
+
+  purge_cruft(starttime);
+
+  DPRINTF(E_DBG, L_LIB, "Running post library scan jobs\n");
+  db_hook_post_scan();
+
+  endtime = time(NULL);
+  DPRINTF(E_LOG, L_LIB, "Library meta rescan completed in %.f sec (%d changes)\n", difftime(endtime, starttime), deferred_update_notifications);
+  scanning = false;
+
+  if (handle_deferred_update_notifications())
+    listener_notify(LISTENER_UPDATE | LISTENER_DATABASE);
+  else
+    listener_notify(LISTENER_UPDATE);
+
+  *ret = 0;
+  return COMMAND_END;
+}
+
+
 static enum command_state
 fullrescan(void *arg, int *ret)
 {
@@ -309,7 +371,7 @@ fullrescan(void *arg, int *ret)
 
   player_playback_stop();
   db_queue_clear(0);
-  db_purge_all(); // Clears files, playlists, playlistitems, inotify and groups
+  db_purge_all(); // Clears files, playlists, playlistitems, inotify and groups, incl RSS
 
   for (i = 0; sources[i]; i++)
     {
@@ -337,9 +399,165 @@ fullrescan(void *arg, int *ret)
   return COMMAND_END;
 }
 
-/*
- * Callback to notify listeners of database changes
- */
+static enum command_state
+playlist_item_add(void *arg, int *retval)
+{
+  struct playlist_item_add_param *param = arg;
+  int i;
+  int ret = LIBRARY_ERROR;
+
+  DPRINTF(E_DBG, L_LIB, "Adding item '%s' to playlist '%s'\n", param->vp_item, param->vp_playlist);
+
+  for (i = 0; sources[i]; i++)
+    {
+      if (sources[i]->disabled || !sources[i]->playlist_item_add)
+	{
+	  DPRINTF(E_DBG, L_LIB, "Library source '%s' is disabled or does not support playlist_item_add\n", sources[i]->name);
+	  continue;
+	}
+
+      ret = sources[i]->playlist_item_add(param->vp_playlist, param->vp_item);
+
+      if (ret == LIBRARY_OK)
+	{
+	  DPRINTF(E_DBG, L_LIB, "Adding item '%s' to playlist '%s' with library source '%s'\n", param->vp_item, param->vp_playlist, sources[i]->name);
+	  listener_notify(LISTENER_STORED_PLAYLIST);
+	  break;
+	}
+    }
+
+  *retval = ret;
+  return COMMAND_END;
+}
+
+static enum command_state
+playlist_remove(void *arg, int *retval)
+{
+  const char *virtual_path = arg;
+  int i;
+  int ret = LIBRARY_ERROR;
+
+  DPRINTF(E_DBG, L_LIB, "Removing playlist at path '%s'\n", virtual_path);
+
+  for (i = 0; sources[i]; i++)
+    {
+      if (sources[i]->disabled || !sources[i]->playlist_remove)
+	{
+	  DPRINTF(E_DBG, L_LIB, "Library source '%s' is disabled or does not support playlist_remove\n", sources[i]->name);
+	  continue;
+	}
+
+      ret = sources[i]->playlist_remove(virtual_path);
+
+      if (ret == LIBRARY_OK)
+	{
+	  DPRINTF(E_DBG, L_LIB, "Removing playlist '%s' with library source '%s'\n", virtual_path, sources[i]->name);
+	  listener_notify(LISTENER_STORED_PLAYLIST);
+	  break;
+	}
+    }
+
+  *retval = ret;
+  return COMMAND_END;
+}
+
+static enum command_state
+queue_item_add(void *arg, int *retval)
+{
+  struct queue_item_add_param *param = arg;
+  int i;
+  int ret;
+
+  DPRINTF(E_DBG, L_LIB, "Add items for path '%s' to the queue\n", param->path);
+
+  ret = LIBRARY_PATH_INVALID;
+  for (i = 0; sources[i] && ret == LIBRARY_PATH_INVALID; i++)
+    {
+      if (sources[i]->disabled || !sources[i]->queue_item_add)
+        {
+	  DPRINTF(E_DBG, L_LIB, "Library source '%s' is disabled or does not support queue_add\n", sources[i]->name);
+	  continue;
+	}
+
+      ret = sources[i]->queue_item_add(param->path, param->position, param->reshuffle, param->item_id, param->count, param->new_item_id);
+
+      if (ret == LIBRARY_OK)
+	{
+	  DPRINTF(E_DBG, L_LIB, "Items for path '%s' from library source '%s' added to the queue\n", param->path, sources[i]->name);
+	  break;
+	}
+    }
+
+  if (ret != LIBRARY_OK)
+    DPRINTF(E_LOG, L_LIB, "Failed to add items for path '%s' to the queue (%d)\n", param->path, ret);
+
+  *retval = ret;
+  return COMMAND_END;
+}
+
+static enum command_state
+queue_save(void *arg, int *retval)
+{
+  const char *virtual_path = arg;
+  int i;
+  int ret = LIBRARY_ERROR;
+
+  DPRINTF(E_DBG, L_LIB, "Saving queue to path '%s'\n", virtual_path);
+
+  for (i = 0; sources[i]; i++)
+    {
+      if (sources[i]->disabled || !sources[i]->queue_save)
+	{
+	  DPRINTF(E_DBG, L_LIB, "Library source '%s' is disabled or does not support queue_save\n", sources[i]->name);
+	  continue;
+	}
+
+      ret = sources[i]->queue_save(virtual_path);
+
+      if (ret == LIBRARY_OK)
+	{
+	  DPRINTF(E_DBG, L_LIB, "Saving queue to path '%s' with library source '%s'\n", virtual_path, sources[i]->name);
+	  listener_notify(LISTENER_STORED_PLAYLIST);
+	  break;
+	}
+    }
+
+  *retval = ret;
+  return COMMAND_END;
+}
+
+static enum command_state
+item_add(void *arg, int *retval)
+{
+  const char *path = arg;
+  int i;
+  int ret = LIBRARY_ERROR;
+
+  DPRINTF(E_DBG, L_LIB, "Adding item to library '%s'\n", path);
+
+  for (i = 0; sources[i]; i++)
+    {
+      if (sources[i]->disabled || !sources[i]->item_add)
+	{
+	  DPRINTF(E_DBG, L_LIB, "Library source '%s' is disabled or does not support add_item\n", sources[i]->name);
+	  continue;
+	}
+
+      ret = sources[i]->item_add(path);
+
+      if (ret == LIBRARY_OK)
+	{
+	  DPRINTF(E_DBG, L_LIB, "Add item to path '%s' with library source '%s'\n", path, sources[i]->name);
+	  listener_notify(LISTENER_DATABASE);
+	  break;
+	}
+    }
+
+  *retval = ret;
+  return COMMAND_END;
+}
+
+// Callback to notify listeners of database changes
 static void
 update_trigger_cb(int fd, short what, void *arg)
 {
@@ -369,7 +587,7 @@ update_trigger(void *arg, int *retval)
 }
 
 
-/* --------------------------- LIBRARY INTERFACE -------------------------- */
+/* ----------------------- LIBRARY EXTERNAL INTERFACE ---------------------- */
 
 void
 library_rescan()
@@ -384,6 +602,19 @@ library_rescan()
   commands_exec_async(cmdbase, rescan, NULL);
 }
 
+void
+library_metarescan()
+{
+  if (scanning)
+    {
+      DPRINTF(E_INFO, L_LIB, "Scan already running, ignoring request to trigger metadata scan\n");
+      return;
+    }
+
+  scanning = true; // TODO Guard "scanning" with a mutex
+  commands_exec_async(cmdbase, metarescan, NULL);
+}
+
 void
 library_fullrescan()
 {
@@ -441,39 +672,24 @@ initscan()
     listener_notify(LISTENER_UPDATE);
 }
 
-/*
- * @return true if scan is running, otherwise false
- */
 bool
 library_is_scanning()
 {
   return scanning;
 }
 
-/*
- * @param is_scanning true if scan is running, otherwise false
- */
 void
 library_set_scanning(bool is_scanning)
 {
   scanning = is_scanning;
 }
 
-/*
- * @return true if a running scan should be aborted due to imminent shutdown, otherwise false
- */
 bool
 library_is_exiting()
 {
   return scan_exit;
 }
 
-/*
- * Trigger for sending the DATABASE event
- *
- * Needs to be called, if an update to the database (library tables) occurred. The DATABASE event
- * is emitted with the delay 'library_update_wait'. It is safe to call this function from any thread.
- */
 void
 library_update_trigger(short update_events)
 {
@@ -494,81 +710,17 @@ library_update_trigger(short update_events)
     }
 }
 
-static enum command_state
-playlist_add(void *arg, int *retval)
-{
-  struct playlist_add_param *param = arg;
-  int i;
-  int ret = LIBRARY_ERROR;
-
-  DPRINTF(E_DBG, L_LIB, "Adding item '%s' to playlist '%s'\n", param->vp_item, param->vp_playlist);
-
-  for (i = 0; sources[i]; i++)
-    {
-      if (sources[i]->disabled || !sources[i]->playlist_add)
-	{
-	  DPRINTF(E_DBG, L_LIB, "Library source '%s' is disabled or does not support playlist_add\n", sources[i]->name);
-	  continue;
-	}
-
-      ret = sources[i]->playlist_add(param->vp_playlist, param->vp_item);
-
-      if (ret == LIBRARY_OK)
-	{
-	  DPRINTF(E_DBG, L_LIB, "Adding item '%s' to playlist '%s' with library source '%s'\n", param->vp_item, param->vp_playlist, sources[i]->name);
-	  listener_notify(LISTENER_STORED_PLAYLIST);
-	  break;
-	}
-    }
-
-  *retval = ret;
-  return COMMAND_END;
-}
-
 int
-library_playlist_add(const char *vp_playlist, const char *vp_item)
+library_playlist_item_add(const char *vp_playlist, const char *vp_item)
 {
-  struct playlist_add_param param;
+  struct playlist_item_add_param param;
 
   if (library_is_scanning())
     return -1;
 
   param.vp_playlist = vp_playlist;
   param.vp_item = vp_item;
-  return commands_exec_sync(cmdbase, playlist_add, NULL, &param);
-}
-
-static enum command_state
-playlist_remove(void *arg, int *retval)
-{
-  const char *virtual_path;
-  int i;
-  int ret = LIBRARY_ERROR;
-
-  virtual_path = arg;
-
-  DPRINTF(E_DBG, L_LIB, "Removing playlist at path '%s'\n", virtual_path);
-
-  for (i = 0; sources[i]; i++)
-    {
-      if (sources[i]->disabled || !sources[i]->playlist_remove)
-	{
-	  DPRINTF(E_DBG, L_LIB, "Library source '%s' is disabled or does not support playlist_remove\n", sources[i]->name);
-	  continue;
-	}
-
-      ret = sources[i]->playlist_remove(virtual_path);
-
-      if (ret == LIBRARY_OK)
-	{
-	  DPRINTF(E_DBG, L_LIB, "Removing playlist '%s' with library source '%s'\n", virtual_path, sources[i]->name);
-	  listener_notify(LISTENER_STORED_PLAYLIST);
-	  break;
-	}
-    }
-
-  *retval = ret;
-  return COMMAND_END;
+  return commands_exec_sync(cmdbase, playlist_item_add, NULL, &param);
 }
 
 int
@@ -580,39 +732,6 @@ library_playlist_remove(char *virtual_path)
   return commands_exec_sync(cmdbase, playlist_remove, NULL, virtual_path);
 }
 
-static enum command_state
-queue_save(void *arg, int *retval)
-{
-  const char *virtual_path;
-  int i;
-  int ret = LIBRARY_ERROR;
-
-  virtual_path = arg;
-
-  DPRINTF(E_DBG, L_LIB, "Saving queue to path '%s'\n", virtual_path);
-
-  for (i = 0; sources[i]; i++)
-    {
-      if (sources[i]->disabled || !sources[i]->queue_save)
-	{
-	  DPRINTF(E_DBG, L_LIB, "Library source '%s' is disabled or does not support queue_save\n", sources[i]->name);
-	  continue;
-	}
-
-      ret = sources[i]->queue_save(virtual_path);
-
-      if (ret == LIBRARY_OK)
-	{
-	  DPRINTF(E_DBG, L_LIB, "Saving queue to path '%s' with library source '%s'\n", virtual_path, sources[i]->name);
-	  listener_notify(LISTENER_STORED_PLAYLIST);
-	  break;
-	}
-    }
-
-  *retval = ret;
-  return COMMAND_END;
-}
-
 int
 library_queue_save(char *path)
 {
@@ -622,15 +741,33 @@ library_queue_save(char *path)
   return commands_exec_sync(cmdbase, queue_save, NULL, path);
 }
 
-/*
- * Execute the function 'func' with the given argument 'arg' in the library thread.
- *
- * The pointer passed as argument is freed in the library thread after func returned.
- *
- * @param func The function to be executed
- * @param arg Argument passed to func
- * @return 0 if triggering the function execution succeeded, -1 on failure.
- */
+int
+library_queue_item_add(const char *path, int position, char reshuffle, uint32_t item_id, int *count, int *new_item_id)
+{
+  struct queue_item_add_param param;
+
+  if (library_is_scanning())
+    return -1;
+
+  param.path = path;
+  param.position = position;
+  param.reshuffle = reshuffle;
+  param.item_id = item_id;
+  param.count = count;
+  param.new_item_id = new_item_id;
+
+  return commands_exec_sync(cmdbase, queue_item_add, NULL, &param);
+}
+
+int
+library_item_add(const char *path)
+{
+  if (library_is_scanning())
+    return -1;
+
+  return commands_exec_sync(cmdbase, item_add, NULL, (char *)path);
+}
+
 int
 library_exec_async(command_function func, void *arg)
 {
@@ -693,12 +830,18 @@ library_init(void)
 
   for (i = 0; sources[i]; i++)
     {
-      if (!sources[i]->init)
-	continue;
+      if (!sources[i]->initscan || !sources[i]->rescan || !sources[i]->metarescan || !sources[i]->fullrescan)
+	{
+	  DPRINTF(E_FATAL, L_LIB, "BUG: library source '%s' is missing a scanning method\n", sources[i]->name);
+	  return -1;
+	}
 
-      ret = sources[i]->init();
-      if (ret < 0)
-	sources[i]->disabled = 1;
+      if (sources[i]->init && !sources[i]->disabled)
+	{
+	  ret = sources[i]->init();
+	  if (ret < 0)
+	    sources[i]->disabled = 1;
+	}
     }
 
   CHECK_NULL(L_LIB, cmdbase = commands_base_new(evbase_lib, NULL));
@@ -735,8 +878,15 @@ library_deinit()
   for (i = 0; sources[i]; i++)
     {
       if (sources[i]->deinit && !sources[i]->disabled)
-      sources[i]->deinit();
+	sources[i]->deinit();
     }
 
+  for (i = 0; i < ARRAY_SIZE(library_cb_register); i++)
+    {
+      if (library_cb_register[i].ev)
+	event_free(library_cb_register[i].ev);
+    }
+
+  event_free(updateev);
   event_base_free(evbase_lib);
 }
