@@ -54,6 +54,7 @@
 #include <libavutil/log.h>
 #include <libavformat/avformat.h>
 #include <libavfilter/avfilter.h>
+#include <curl/curl.h>
 
 #include <pthread.h>
 #include <gcrypt.h>
@@ -75,10 +76,6 @@ GCRY_THREAD_OPTION_PTHREAD_IMPL;
 # include "lastfm.h"
 #endif
 
-#ifdef HAVE_LIBCURL
-# include <curl/curl.h>
-#endif
-
 #define PIDFILE   STATEDIR "/run/" PACKAGE ".pid"
 #define WEB_ROOT  DATADIR "/htdocs"
 
@@ -313,7 +310,7 @@ register_services(char *ffid, bool no_web, bool no_rsp, bool no_daap, bool no_mp
   // otherwise at least my Sony speaker won't use the service.
 
   // Use as scratch space for the hash
-  snprintf(records[4], 128, "iTunes_Ctrl_%" PRIX64, libhash);
+  snprintf(records[4], 128, "iTunes_Ctrl_%016" PRIX64, libhash);
 
   ret = mdns_register(records[4], "_dacp._tcp", port, txtrecord);
   if (ret < 0)
@@ -446,7 +443,7 @@ signal_kqueue_cb(int fd, short event, void *arg)
 }
 #endif
 
-
+#if (LIBAVCODEC_VERSION_MAJOR < 58) || ((LIBAVCODEC_VERSION_MAJOR == 58) && (LIBAVCODEC_VERSION_MINOR < 18))
 static int
 ffmpeg_lockmgr(void **pmutex, enum AVLockOp op)
 {
@@ -477,6 +474,7 @@ ffmpeg_lockmgr(void **pmutex, enum AVLockOp op)
 
   return 1;
 }
+#endif
 
 int
 main(int argc, char **argv)
@@ -660,6 +658,8 @@ main(int argc, char **argv)
   DPRINTF(E_INFO, L_MAIN, "Initialized with libav %s\n", av_version);
 #endif
 
+// The following was deprecated with ffmpeg 4.0 = avcodec 58.18, avformat 58.12, avfilter 7.16
+#if (LIBAVCODEC_VERSION_MAJOR < 58) || ((LIBAVCODEC_VERSION_MAJOR == 58) && (LIBAVCODEC_VERSION_MINOR < 18))
   ret = av_lockmgr_register(ffmpeg_lockmgr);
   if (ret < 0)
     {
@@ -668,18 +668,21 @@ main(int argc, char **argv)
       ret = EXIT_FAILURE;
       goto ffmpeg_init_fail;
     }
-
+#endif
+#if (LIBAVFORMAT_VERSION_MAJOR < 58) || ((LIBAVFORMAT_VERSION_MAJOR == 58) && (LIBAVFORMAT_VERSION_MINOR < 12))
   av_register_all();
+#endif
+#if (LIBAVFILTER_VERSION_MAJOR < 7) || ((LIBAVFILTER_VERSION_MAJOR == 7) && (LIBAVFILTER_VERSION_MINOR < 16))
   avfilter_register_all();
+#endif
+
 #if HAVE_DECL_AVFORMAT_NETWORK_INIT
   avformat_network_init();
 #endif
   av_log_set_callback(logger_ffmpeg);
 
-#ifdef HAVE_LIBCURL
   /* Initialize libcurl */
   curl_global_init(CURL_GLOBAL_DEFAULT);
-#endif
 
   /* Initialize libgcrypt */
   gcry_control(GCRYCTL_SET_THREAD_CBS, &gcry_threads_pthread);
@@ -911,6 +914,8 @@ main(int argc, char **argv)
   DPRINTF(E_LOG, L_MAIN, "Stopping gracefully\n");
   ret = EXIT_SUCCESS;
 
+  event_free(sig_event);
+
   /*
    * On a clean shutdown, bring mDNS down first to give a chance
    * to the clients to perform a clean shutdown on their end
@@ -963,6 +968,8 @@ main(int argc, char **argv)
     }
 
  mdns_fail:
+  event_base_free(evbase_main);
+
  daemon_fail:
   if (background)
     {
@@ -979,15 +986,16 @@ main(int argc, char **argv)
 
  signal_block_fail:
  gcrypt_init_fail:
-#ifdef HAVE_LIBCURL
   curl_global_cleanup();
-#endif
 #if HAVE_DECL_AVFORMAT_NETWORK_INIT
   avformat_network_deinit();
 #endif
-  av_lockmgr_register(NULL);
 
+#if (LIBAVCODEC_VERSION_MAJOR < 58) || ((LIBAVCODEC_VERSION_MAJOR == 58) && (LIBAVCODEC_VERSION_MINOR < 18))
+  av_lockmgr_register(NULL);
  ffmpeg_init_fail:
+#endif
+
   DPRINTF(E_LOG, L_MAIN, "Exiting.\n");
   conffile_unload();
   logger_deinit();
