@@ -718,8 +718,7 @@ playback_eot(void *arg, int *retval)
 
   g_state = SPOTIFY_STATE_STOPPING;
 
-  // TODO 1) This will block for a while, but perhaps ok?
-  input_write(spotify_audio_buffer, INPUT_FLAG_EOF);
+  input_write(spotify_audio_buffer, NULL, INPUT_FLAG_EOF);
 
   *retval = 0;
   return COMMAND_END;
@@ -1007,17 +1006,22 @@ logged_out(sp_session *sess)
 static int music_delivery(sp_session *sess, const sp_audioformat *format,
                           const void *frames, int num_frames)
 {
+  struct media_quality quality = { 0 };
   size_t size;
   int ret;
 
   /* No support for resampling right now */
-  if ((format->sample_rate != 44100) || (format->channels != 2))
+  if ((format->sample_type != SP_SAMPLETYPE_INT16_NATIVE_ENDIAN) || (format->channels != 2))
     {
-      DPRINTF(E_LOG, L_SPOTIFY, "Got music with unsupported samplerate or channels, stopping playback\n");
+      DPRINTF(E_LOG, L_SPOTIFY, "Got music with unsupported sample format or number of channels, stopping playback\n");
       spotify_playback_stop_nonblock();
       return num_frames;
     }
 
+  quality.sample_rate = format->sample_rate;
+  quality.bits_per_sample = 16;
+  quality.channels = format->channels;
+
   // Audio discontinuity, e.g. seek
   if (num_frames == 0)
     {
@@ -1037,7 +1041,7 @@ static int music_delivery(sp_session *sess, const sp_audioformat *format,
   // The input buffer only accepts writing when it is approaching depletion, and
   // because we use NONBLOCK it will just return if this is not the case. So in
   // most cases no actual write is made and spotify_audio_buffer will just grow.
-  input_write(spotify_audio_buffer, INPUT_FLAG_NONBLOCK);
+  input_write(spotify_audio_buffer, &quality, 0);
 
   return num_frames;
 }
@@ -1336,6 +1340,40 @@ spotify_status_info_get(struct spotify_status_info *info)
   CHECK_ERR(L_SPOTIFY, pthread_mutex_unlock(&status_lck));
 }
 
+/* Thread: library, httpd */
+static int
+logout(char **errmsg)
+{
+  sp_error err;
+
+  CHECK_ERR(L_SPOTIFY, pthread_mutex_lock(&login_lck));
+
+  if (SP_CONNECTION_STATE_LOGGED_IN != fptr_sp_session_connectionstate(g_sess))
+    {
+      CHECK_ERR(L_SPOTIFY, pthread_mutex_unlock(&login_lck));
+      return 0;
+    }
+
+  DPRINTF(E_LOG, L_SPOTIFY, "Logging out of Spotify (current state is %d)\n", g_state);
+
+  fptr_sp_session_player_unload(g_sess);
+  err = fptr_sp_session_logout(g_sess);
+  if (SP_ERROR_OK != err)
+    {
+      DPRINTF(E_LOG, L_SPOTIFY, "Could not logout of Spotify: %s\n", fptr_sp_error_message(err));
+      if (errmsg)
+	*errmsg = safe_asprintf("Could not logout of Spotify: %s", fptr_sp_error_message(err));
+
+      CHECK_ERR(L_SPOTIFY, pthread_mutex_unlock(&login_lck));
+      return -1;
+    }
+
+  CHECK_ERR(L_SPOTIFY, pthread_cond_wait(&login_cond, &login_lck)); // Wait for logged_out()
+  CHECK_ERR(L_SPOTIFY, pthread_mutex_unlock(&login_lck));
+
+  return 0;
+}
+
 /* Thread: library, httpd */
 static int
 login_user(const char *user, const char *password, char **errmsg)
@@ -1361,28 +1399,9 @@ login_user(const char *user, const char *password, char **errmsg)
       return -1;
     }
 
-  if (SP_CONNECTION_STATE_LOGGED_IN == fptr_sp_session_connectionstate(g_sess))
-    {
-      CHECK_ERR(L_SPOTIFY, pthread_mutex_lock(&login_lck));
-
-      DPRINTF(E_LOG, L_SPOTIFY, "Logging out of Spotify (current state is %d)\n", g_state);
-
-      fptr_sp_session_player_unload(g_sess);
-      err = fptr_sp_session_logout(g_sess);
-
-      if (SP_ERROR_OK != err)
-	{
-	  DPRINTF(E_LOG, L_SPOTIFY, "Could not logout of Spotify: %s\n", fptr_sp_error_message(err));
-	  if (errmsg)
-	    *errmsg = safe_asprintf("Could not logout of Spotify: %s", fptr_sp_error_message(err));
-
-	  CHECK_ERR(L_SPOTIFY, pthread_mutex_unlock(&login_lck));
-	  return -1;
-	}
-
-      CHECK_ERR(L_SPOTIFY, pthread_cond_wait(&login_cond, &login_lck));
-      CHECK_ERR(L_SPOTIFY, pthread_mutex_unlock(&login_lck));
-    }
+  ret = logout(errmsg);
+  if (ret < 0)
+    return -1;
 
   CHECK_ERR(L_SPOTIFY, pthread_mutex_lock(&login_lck));
 
@@ -1453,6 +1472,14 @@ spotify_login(char **arglist)
     spotify_login_user(NULL, NULL, NULL);
 }
 
+void
+spotify_logout(void)
+{
+  logout(NULL);
+
+  spotifywebapi_purge();
+}
+
 /* Thread: main */
 int
 spotify_init(void)
@@ -1578,6 +1605,7 @@ spotify_init(void)
   
  session_fail:
  cmd_fail:
+  event_free(g_notifyev);
  evnew_fail:
   commands_base_free(cmdbase);
   event_base_free(evbase_spotify);
@@ -1621,7 +1649,8 @@ spotify_deinit(void)
   /* Release session */
   fptr_sp_session_release(g_sess);
 
-  /* Free event base (should free events too) */
+  /* Free event base */
+  event_free(g_notifyev);
   event_base_free(evbase_spotify);
 
   /* Close pipes */
